#if !defined HATP_TASK_INSTANCE_HH_
#define HATP_TASK_INSTANCE_HH_

#include<vector>
#include<string>
#include<map>

#include "Task.hh"

class TaskList;
class HTNBase;
class LightWorldBase;
class WorldBase;
class WorldBaseReport;

class TaskInstance {

private:
	/**
	* The associated task
	*/
	Task const& generalTask;

	/**
	* vector of the arguments of the task
	*/
	std::vector<std::string> args;

	/**
	* minimum Time for action realization
	*/
	double minTime;

	/**
	* maximum Time for action realization
	*/
	double maxTime;

public:

	/**
	* Pointer to stock of general tasks
	*/
	static HTNBase* htn_base;

	explicit TaskInstance(std::string const& taskName, std::vector<std::string> const& vecArgs, LightWorldBase const& wb);
	explicit TaskInstance(std::string const& taskName, std::vector<std::string> const& vecArgs, WorldBase const& wb);
	explicit TaskInstance(std::string const& taskName, std::vector<std::string> const& vecArgs, std::map<std::string, std::string> const& varAndTypes);
	TaskInstance(TaskInstance const& ti);
	Task const& getGeneralTask() const; 
	std::vector<std::string> const& getArgs() const;
	void setHTNBase(HTNBase* argHTNBase);
	bool checkPreconditions(LightWorldBase const& wb) const;
	bool checkPreconditionsWithReport(LightWorldBase const& wb, WorldBaseReport& report) const;
	bool checkTaskEmpty(LightWorldBase const& wb) const;
	std::vector<TaskList*> getDecompositions(LightWorldBase const& wb, HTNBase const& htnb) const;
	void applyActionEffects(LightWorldBase& wb) const;
	void applyActionEffectsWithReport(LightWorldBase& wb, WorldBaseReport& report) const;
	float ComputeCost(LightWorldBase const& wb) const;
	void ComputeTimeLimit(LightWorldBase const& wb);
	double getMinTime() const;
	double getMaxTime() const;
	std::string toString() const;
	std::vector<std::string> getAssociatedAgents() const;
	bool isAction() const;
	bool operator==(TaskInstance const& ti) const;
	~TaskInstance();
};

/**
 * constructor by copying
 */
inline TaskInstance::TaskInstance(TaskInstance const& ti)
	: generalTask(ti.getGeneralTask()), args(ti.getArgs()), minTime(ti.getMinTime()),
	maxTime(ti.getMaxTime()) {}

/**
 * return generalTask
 */
inline Task const& TaskInstance::getGeneralTask() const {
	return generalTask;
}

/**
 * return args
 */
inline std::vector<std::string> const& TaskInstance::getArgs() const {
	return args;
}

/**
 * return generalTask->isAction()
 */
inline bool TaskInstance::isAction() const {
	return generalTask.isAction();
}

/**
 * return minTime
 */
inline double TaskInstance::getMinTime() const {
	return minTime;
}

/**
 * return maxTime
 */
inline double TaskInstance::getMaxTime() const {
	return maxTime;
}

/**
 * destructor: empty
 */
inline TaskInstance::~TaskInstance() {}

#endif
