#include<iostream>
#include<sstream>
#include <fstream>
#include<string>
#include<vector>

#include "hatpPlanningSession.hh"
#include "wait.hh"

using namespace std;
using namespace boost::posix_time;

unsigned long int rejectCounter=0;
unsigned long int choiceCounter=0;
unsigned long int planCounter=0;

/**
 * constructor: sets wb to argWB, htnb to argHTN, current_TL to TL an sets choice_tree to a new ChoiceReminder make from TL
 */
hatpPlanningSession::hatpPlanningSession(WorldBase const& argWB, HTNBase const& argHTN, SocialRuleBase const& argSRB, TaskList const& TL, 
	PartialPlanAppraiser const& ppa, PlanSender<MainSessionMsg>* stock, bool argTimeLimit, unsigned long int argDuration)
: htnb(argHTN), srb(argSRB), original_WB(new WorldBase(argWB)), choice_tree(ChoiceReminder(argWB, TL, ppa)),
current_WB(choice_tree.getCurrentWorldBase()), current_TL(choice_tree.getCurrentTaskList()),
current_PT(choice_tree.getCurrentPlanTree()), current_Eval(choice_tree.getCurrentPartialEval()), communicator(stock) {
	current_T = EMPTY_ID;
	current_projection_order = vector<unsigned long int>(0);
	time_limit = argTimeLimit;
	duration = argDuration;
	current_Eval->updateBadDecompoFields(srb, *current_WB, *current_PT);
}

/**
 * constructor: used essentially for partial replanning sets arguments to their corresponding fields, and sets choice_tree to a
 * new ChoiceReminder(). If PT==NULL same effect as constructor without PlanTree arguments
 */
hatpPlanningSession::hatpPlanningSession(WorldBase const& argWB, HTNBase const& argHTN, SocialRuleBase const& argSRB, TaskList const& TL, 
	PlanTree const& PT, PartialPlanAppraiser const& ppa, PlanSender<MainSessionMsg>* stock, bool argTimeLimit, unsigned long int argDuration): 
htnb(argHTN), srb(argSRB), original_WB(new WorldBase(argWB)), choice_tree(ChoiceReminder(new ChoicePoint(ROOT_TYPE, argWB, TL, 
vector<unsigned long int>(0), PT, ppa))), current_WB(choice_tree.getCurrentWorldBase()), current_TL(choice_tree.getCurrentTaskList()), 
current_PT(choice_tree.getCurrentPlanTree()), current_Eval(choice_tree.getCurrentPartialEval()), communicator(stock) {
	current_T = EMPTY_ID;
	current_projection_order = choice_tree.getCurrentProjectionOrder();
	time_limit = argTimeLimit;
	duration = argDuration;
	current_Eval->updateBadDecompoFields(srb, *current_WB, *current_PT);
}

/**
 * constructor by copying: needed by boost thread to prevent segmentation fault
 */
hatpPlanningSession::hatpPlanningSession(hatpPlanningSession const& otherSession) : htnb(otherSession.getHTNBase()), srb(otherSession.getSocialRuleBase()),
original_WB(otherSession.getOriginalWorldBase()), choice_tree(otherSession.getChoiceReminder()), current_WB(choice_tree.getCurrentWorldBase()), 
current_TL(choice_tree.getCurrentTaskList()), current_projection_order(choice_tree.getCurrentProjectionOrder()), current_PT(choice_tree.getCurrentPlanTree()),
current_Eval(choice_tree.getCurrentPartialEval()), time_limit(otherSession.getTimeLimit()), duration(otherSession.getDuration()), 
communicator(otherSession.getCommunicator()) {}

/**
 * runs the planning session during duration if time_limit is true (duration is expressed in ms), all found plans will be written
 * in files specified by constant HEAD_PLAN_FILE with numbers
 */
void hatpPlanningSession::run() {
	string s;
	vector<ChoicePoint*> vecCP;
	vector<unsigned long int> vecFirstTasks, vecFirstUnlockedTasks;
	unsigned long int nbFirstTasks, nbFirstUnlockedTasks;
	vector<TaskList*> current_TL_copies;
	vector<PlanTree*> current_PT_copies;
	size_t i,j;
	bool DEBUG_ON = false;
	planCounter = 0;
	ptime start = microsec_clock::local_time();
	ptime estimated_end = start + microsec(duration*1000);
	bool allExplored = false, emptyTasks;
	map<string, float> bestScoreDetails;
	bool newPossibleBestScore = false, waitingAnAnswer = false;
	LightWorldBase lwb(*original_WB);
	
	// start run main loop
	while( !allExplored && (!time_limit || (time_limit && microsec_clock::local_time() < estimated_end)) ) {
		if(newPossibleBestScore) { // we get best plan data if there are new ones
			if(!waitingAnAnswer) { // to prevent sending too much messages (in case a previous answer has not been received yet)
				communicator->sendObjectToA("", GetBestScore);
				waitingAnAnswer = true;
			}
			// we wait an answer with a deadline of 1ms to save time (empirical value)
			if(waitBySelect(communicator->fdToWaitA(), 0.001).first && communicator->typeOfMessageFromA() == BestScore && 
					communicator->objectAvailableFromA()) {
				bestScoreDetails = *(communicator->receiveObjectFromA<map<string, float> const*>());
				newPossibleBestScore = false;
				waitingAnAnswer = false;
			}
		}
		// we apply branch and bound if we have a reference plan
		if(!bestScoreDetails.empty()) {
			float contribution = this->computeContribution(bestScoreDetails, *current_Eval);
			if(contribution > 0.) {
				choice_tree.setCurrentPointTag(COST_EXCEEDED_TAG);
				allExplored = choice_tree.backtrack(); 
				this->updateCurrentData(); 
			}
		}
		// end branch and bound optimization
		if(!allExplored) { 
			do { // we reduce all firsts TODO tasks which have their empty conditions verified to empty tasks
				emptyTasks = false;
				vecFirstTasks = current_TL->getFirstTasks(*current_WB);
				for(i=0 ; i<vecFirstTasks.size() ; ++i) {
					if( current_TL->find(vecFirstTasks[i])->getStatus() == TODO && 
							current_TL->find(vecFirstTasks[i])->getInstance().checkTaskEmpty(*current_WB) ) {
						current_TL->substitute(vecFirstTasks[i], TaskList());
						current_PT->decompose(vecFirstTasks[i], TaskList());
						emptyTasks = true;
					}
				}
			} while(emptyTasks == true);
			nbFirstTasks = vecFirstTasks.size();
			if( nbFirstTasks == 0 ) { 
				if(current_TL->allDone()) { 
					if(current_TL->isEmpty()) {
						allExplored=true;
						communicator->sendObjectToA("", SolFound);
					}
					else {
						planCounter++;
						Plan* p = new Plan(*current_TL, current_projection_order, *current_PT, *current_Eval);
						if(p->postTreatments(lwb, srb, -1.) ){
							communicator->sendObjectToA(p, NewPlan);
							newPossibleBestScore = true;
						}
						else delete p;
						choice_tree.setCurrentPointTag(SUCCESS_TAG);
						allExplored = choice_tree.backtrack(); 
						this->updateCurrentData(); 
					}
				}
				else { 
					rejectCounter+=1;
					choice_tree.setCurrentPointTag(FAILED_TAG);
					allExplored = choice_tree.backtrack(); 
					this->updateCurrentData(); 
				}
			}
			else if( nbFirstTasks == 1 ) { 
				if(current_TL->find(vecFirstTasks[0])->getStatus() == TODO_LOCKED) {
					current_TL->find(vecFirstTasks[0])->setStatus(TODO);
				}
				current_T = vecFirstTasks[0];
				allExplored = this->runPhase2();
			}
			else { 
				if(choice_tree.getCurrentPointType() == PARALLEL_TYPE && choice_tree.getCurrentFatherChoices() == vecFirstTasks) 
					// the new parallel choice point is not necessary
					allExplored = this->runPhase2();
				else { 
					vecFirstUnlockedTasks = current_TL->getFirstUnlockedTasks(*current_WB);
					nbFirstUnlockedTasks = vecFirstUnlockedTasks.size();
					if(nbFirstUnlockedTasks==0) { 
						current_TL->unlockAll();
						vecCP = vector<ChoicePoint*>(nbFirstTasks);
						current_TL_copies = current_TL->getMultipleCopies(nbFirstTasks);
						current_PT_copies = current_PT->getMultipleCopies(nbFirstTasks);
						for(i=0 ; i<nbFirstTasks ; ++i) {
							for(j=0 ; j<nbFirstTasks ; ++j) 
								if(j != i) current_TL_copies[i]->find(vecFirstTasks[j])->setStatus(TODO_LOCKED);
							vecCP[i] = new ChoicePoint(PARALLEL_TYPE, *current_WB, current_TL_copies[i], current_projection_order, current_PT_copies[i], vecFirstTasks[i], *current_Eval);
						}
						choiceCounter+= vecCP.size();
						choice_tree.addChoicesToCurrentNode(vecCP);
						choice_tree.selectUnexploredChildOfCurrentPoint();
						this->updateCurrentData(); 
						if(current_T == EMPTY_ID) {
							cerr << "ERROR: when choosing an unexplored branch of a parallelChoicePoint" << endl;
						}
						else allExplored = this->runPhase2();
					}
					else if(nbFirstUnlockedTasks==1) { 
						current_T = vecFirstUnlockedTasks[0];
						allExplored = this->runPhase2();
					}
					else { 
						vecCP = vector<ChoicePoint*>(nbFirstUnlockedTasks);
						current_TL_copies = current_TL->getMultipleCopies(nbFirstUnlockedTasks);
						current_PT_copies = current_PT->getMultipleCopies(nbFirstUnlockedTasks);
						for(i=0 ; i<nbFirstUnlockedTasks ; ++i) {
							for(j=0 ; j<nbFirstUnlockedTasks ; ++j) 
								if(j != i) current_TL_copies[i]->find(vecFirstUnlockedTasks[j])->setStatus(TODO_LOCKED);
							vecCP[i] = new ChoicePoint(PARALLEL_TYPE, *current_WB, current_TL_copies[i], current_projection_order, current_PT_copies[i], vecFirstUnlockedTasks[i], *current_Eval);
						}
						choiceCounter+= vecCP.size();
						choice_tree.addChoicesToCurrentNode(vecCP);
						choice_tree.selectUnexploredChildOfCurrentPoint();
						this->updateCurrentData(); 

						if(current_T == EMPTY_ID) {
							cerr << "ERROR: when choosing an unexplored branch of a parallelChoicePoint" << endl;
						}
						else allExplored = this->runPhase2();
					} // end vecFirstUnlockedTasks 
				} // end creation of a new parallel choice point
			} // end else parallel choice
		} // end if !allExplored
	} // end run main loop
	if(DEBUG_ON) {
		cout << "Planning process stopped after "<< (microsec_clock::local_time() - start).total_milliseconds() << "ms." << endl;
		cout << "Total plans found: "<< planCounter << endl;
	}
	communicator->sendObjectToA("", End);
}

/**
 * the 2nd phase of the planning process: where parallelChoicePoint issued have been treated
 * and a Task is selected in current_T: we work on this task (i.e. on current_T)
 */
bool hatpPlanningSession::runPhase2() {
	bool result;
	TaskInstance& tsk = current_TL->find(current_T)->getInstance();
	size_t i;
	
	if(!tsk.isAction()) { // the task is compound
		// ---------- Belief Management ----------
		if(current_WB->isActiveBM()) {
			// test of each entity in args ...
			vector<string> args = tsk.getArgs();
			vector<TaskListNode*> commTL;
			map<string, LightWorldBaseAgent*> wbas = current_WB->getWBAgents();
			bool infoGiven;
			for(size_t i = 0 ; i < args.size() ; i++) {
				if(current_WB->find(args[i])->isUnknown()) {
					for(map<string, LightWorldBaseAgent*>::iterator it = wbas.begin() ; it != wbas.end() ; it++) {
						if(it->second->containEntity(args[i]) && it->second->find(args[i])->isKnown()) {
							vector<string> commArgs;
							commArgs.push_back(current_WB->getMyselfName());
							commArgs.push_back(it->first);
							commArgs.push_back(args[i]);
							commArgs.push_back(ATTRNULL);
							Task* commTask = htnb.getBeliefMethod()->getTaskOfType(QUESTION);
							commTask->setCommunicationTask(true);
							TaskInstance* commAction = new TaskInstance(commTask->getName(), commArgs, *current_WB);
							TaskListNode* commTLNode = new TaskListNode(*commAction);
							commTL.push_back(commTLNode);
						}
					}
				}
				else {
					// test of each attributes of this arg...
					set<string> argAttrs = current_WB->find(args[i])->getUnknownAttrs();
					for(set<string>::const_iterator itAttr = argAttrs.begin() ; itAttr != argAttrs.end() ; itAttr++) {
						infoGiven = false;
						for(map<string, LightWorldBaseAgent*>::iterator it = wbas.begin() ; it != wbas.end() ; it++) {
							if(it->second->containEntity(args[i]) && it->second->find(args[i])->getKnownAttrs().find(*itAttr) 
							 != it->second->find(args[i])->getKnownAttrs().end() && !infoGiven) {
								vector<string> commArgs;
								commArgs.push_back(current_WB->getMyselfName());
								commArgs.push_back(it->first);
								commArgs.push_back(args[i]);
								commArgs.push_back(*itAttr);
								Task* commTask = htnb.getBeliefMethod()->getTaskOfType(QUESTION);
								commTask->setCommunicationTask(true);
								TaskInstance* commAction = new TaskInstance(commTask->getName(), commArgs, *current_WB);
								TaskListNode* commTLNode = new TaskListNode(*commAction);
								commTL.push_back(commTLNode);
								infoGiven = true;
							}
						}
					}
				}
			}
			if(commTL.size() > 0) {
				// clean TL and add commActions ...
				current_TL = new TaskList();
				current_TL->addNodes(commTL);
				return false;
			}
		}
		// ---------- Method decomposition ----------
		vector<TaskList*> vecDecompo = tsk.getDecompositions(*current_WB, htnb);
		size_t nbDecompos = vecDecompo.size();
		if(nbDecompos == 0) { 
			rejectCounter+=1; 
			choice_tree.setCurrentPointTag(FAILED_TAG);
			result = choice_tree.backtrack(); 
			this->updateCurrentData(); 
			return result;
		}
		else if(nbDecompos == 1) {
			current_PT->decompose(current_T, *vecDecompo[0]);
			current_Eval->updateBadDecompoFields(*vecDecompo[0], srb, *current_WB, *current_PT);
			current_TL->substitute(current_T, vecDecompo[0]); 
			vecDecompo[0]->noNeedToDeleteNodes();
			delete vecDecompo[0];
			return false;
		}
		else { 
			vector<ChoicePoint*> vecCP(nbDecompos);
			vector<TaskList*> current_TL_copies = current_TL->getMultipleCopies(nbDecompos);
			vector<PlanTree*> current_PT_copies = current_PT->getMultipleCopies(nbDecompos);
			choiceCounter+=nbDecompos; 
			for(i=0 ; i<nbDecompos ; ++i) {
				current_PT_copies[i]->decompose(current_T, *vecDecompo[i]);
				current_Eval->updateBadDecompoFields(*vecDecompo[i], srb, *current_WB, *current_PT_copies[i]);
				current_TL_copies[i]->substitute(current_T, vecDecompo[i]);
				vecCP[i] = new ChoicePoint(DECOMPO_TYPE, *current_WB, current_TL_copies[i], current_projection_order, current_PT_copies[i], *current_Eval);
				vecDecompo[i]->noNeedToDeleteNodes();
				delete vecDecompo[i];
			}
			choice_tree.addChoicesToCurrentNode(vecCP);
			choice_tree.selectUnexploredChildOfCurrentPoint();
			this->updateCurrentData();
			return false;
		}
	}
	else { // the task is an action
		// ---------- Belief Management ----------
		if(current_WB->isActiveBM() && !tsk.getGeneralTask().isCommunicationTask()) {
			bool needCommunication = false;
			bool notForAttribute = false;
			vector<string> communicationInfo (4);
			vector<string> agsInTsk = tsk.getAssociatedAgents();
			vector<string> argOfTsk = tsk.getArgs();
			vector<string> precondsReport, data;
			WorldBaseReport report;
			tsk.checkPreconditionsWithReport(*current_WB, report);
			precondsReport = report.getTestsReport();
			// Test if entity from args is unknown
			for(size_t i = 0 ; i < argOfTsk.size() ; i++) {
				for(size_t j = 0 ; j < agsInTsk.size() ; j++) {
					if(agsInTsk[j] != current_WB->getMyselfName()) {
						LightEntity* eTemp = current_WB->findForAgent(agsInTsk[j], argOfTsk[i]);
						if(eTemp != NULL && eTemp->isUnknown()) {
							needCommunication = true;
							communicationInfo[0] = INFORMATION; communicationInfo[1] = agsInTsk[j];
							communicationInfo[2] = argOfTsk[i]; communicationInfo[3] = ATTRNULL;
							notForAttribute = true;
						}
					}
				}
			}
			// Test of each attribute in action precondition
			if(!notForAttribute) {
			for(vector<string>::iterator it = precondsReport.begin() ; it != precondsReport.end() ; it++) {
				data = parseTestReport(*it);
				for(size_t i = 0 ; i < agsInTsk.size() ; i++) {
					LightEntity* eTemp = current_WB->findForAgent(agsInTsk[i], data[1]);
					if(eTemp != NULL && agsInTsk[i] != current_WB->getMyselfName()) {
						// Test if attribute is unknown
						if(eTemp->isAttrUnknown(data[2])) {
							needCommunication = true;
							communicationInfo[0] = INFORMATION; communicationInfo[1] = agsInTsk[i];
							communicationInfo[2] = data[1]; communicationInfo[3] = data[2];
						}
						// Test if attribute value is different from myself value
						if(eTemp->hasAttributeOfName(data[2])) {
							bool needContrComm = false; 
							LightEntity* eMyself = current_WB->find(data[1]);
							if(data[0] == ATOM_TEST_EQUAL || data[0] == ATOM_TEST_EQUAL) {
								if((*eTemp)[data[2]] != (*eMyself)[data[2]]) needContrComm = true;
							}
							else if(data[0] == VECTOR_TEST_IN || data[0] == VECTOR_TEST_NOT_IN) {
								set<string> valTemp = (*eTemp)(data[2]);
								set<string> valMyself = (*eMyself)(data[2]);
								bool isDiff = false;
								if(valTemp.size() != valMyself.size()) isDiff = true;
								for(set<string>::iterator itx = valMyself.begin() ; itx != valMyself.end() ; itx++) {
									 if(find(valTemp.begin(), valTemp.end(), *itx) == valTemp.end()) isDiff = true;
								}
								if(isDiff) needContrComm = true;
							}
							if(needContrComm) {
								needCommunication = true;
								communicationInfo[0] = CONTRADICTION; communicationInfo[1] = agsInTsk[i];
								communicationInfo[2] = data[1]; communicationInfo[3] = data[2];
							}
						}
					}
				} // end for agent
			}} // end for precond
			if(needCommunication) {
				vector<string> commArgs;
				commArgs.push_back(current_WB->getMyselfName());
				commArgs.push_back(communicationInfo[1]);
				commArgs.push_back(communicationInfo[2]);
				commArgs.push_back(communicationInfo[3]);
				Task* commTask = htnb.getBeliefMethod()->getTaskOfType(communicationInfo[0]);
				commTask->setCommunicationTask(true);
				TaskInstance* commAction = new TaskInstance(commTask->getName(), commArgs, *current_WB);
				TaskListNode* commTLNode = new TaskListNode(*commAction);
				commTLNode->AddSuccessor(current_TL->find(current_T), LOW_LEVEL_LINK);
				vector<TaskListNode*> commTL;
				commTL.push_back(commTLNode);
				current_TL->addNodes(commTL);
				return false;
			}
		}
		// ---------------------------------------------
		// --- Apply action
		float cost = tsk.ComputeCost(*current_WB);
		current_TL->project(current_T, cost);
		current_projection_order.push_back(current_T);
		current_Eval->addToCurrentProjectionCost(cost);
		// --- Need to update WBAgent according to action effects
		if(current_WB->isActiveBM() && !tsk.getGeneralTask().isCommunicationTask()) {
			vector<string> agsInTsk = tsk.getAssociatedAgents();
			vector<string> effectsReport, data;
			WorldBaseReport report;
			tsk.applyActionEffectsWithReport(*current_WB, report); // apply effects on myself belief
			effectsReport = report.getModificationsReport();
			for(vector<string>::iterator it = effectsReport.begin() ; it != effectsReport.end() ; it++) {
				data = parseModificationReport(*it);
				for(size_t i = 0 ; i < agsInTsk.size() ; i++) {
					LightEntity* eTemp = current_WB->findForAgent(agsInTsk[i], data[1]);
					if(eTemp != NULL && agsInTsk[i] != current_WB->getMyselfName()) {
						current_WB->updateForAgent(current_WB->getMyselfName(), agsInTsk[i], data[1], data[2]);
					}
				}
			}
		}
		// -------------------------------------------
		else tsk.applyActionEffects(*current_WB);
		tsk.ComputeTimeLimit(*current_WB);
		return false;
	}
	return false;
}


/**
 * function "main" of the thread planning session
 */
void hatpPlanningSession::operator()() {
	this->run();
}

/**
 * destructor: empty
 */
hatpPlanningSession::~hatpPlanningSession() {}

